@@ -127,6 +127,13 @@ enum quiche_error {
 
     // The peer sent more data in CRYPTO frames than we can buffer.
     QUICHE_ERR_CRYPTO_BUFFER_EXCEEDED = -20,
+
+    // The peer sent an ACK frame with an invalid range.
+    QUICHE_ERR_INVALID_ACK_RANGE = -21,
+
+    // The peer send an ACK frame for a skipped packet used for Optimistic ACK
+    // mitigation.
+    QUICHE_ERR_OPTIMISTIC_ACK_DETECTED = -22,
 };
 
 // Returns a human readable string with the quiche version number.
@@ -333,6 +340,10 @@ void quiche_conn_set_qlog_fd(quiche_conn *conn, int fd, const char *log_title,
 // Configures the given session for resumption.
 int quiche_conn_set_session(quiche_conn *conn, const uint8_t *buf, size_t buf_len);
 
+// Sets the `max_idle_timeout` transport parameter, in milliseconds, default is
+// no timeout.
+int quiche_conn_set_max_idle_timeout(quiche_conn *conn, uint64_t v);
+
 typedef struct {
     // The remote address the packet was received from.
     struct sockaddr *from;
@@ -382,14 +393,14 @@ size_t quiche_conn_send_quantum_on_path(const quiche_conn *conn,
 
 // Reads contiguous data from a stream.
 // out_error_code is only set when STREAM_STOPPED or STREAM_RESET are returned.
-// Set to the reported error code associated with STOP_SENDING or STREAM_RESET. 
+// Set to the reported error code associated with STOP_SENDING or STREAM_RESET.
 ssize_t quiche_conn_stream_recv(quiche_conn *conn, uint64_t stream_id,
                                 uint8_t *out, size_t buf_len, bool *fin,
                                 uint64_t *out_error_code);
 
 // Writes data to a stream.
 // out_error_code is only set when STREAM_STOPPED or STREAM_RESET are returned.
-// Set to the reported error code associated with STOP_SENDING or STREAM_RESET. 
+// Set to the reported error code associated with STOP_SENDING or STREAM_RESET.
 ssize_t quiche_conn_stream_send(quiche_conn *conn, uint64_t stream_id,
                                 const uint8_t *buf, size_t buf_len, bool fin,
                                 uint64_t *out_error_code);
@@ -409,7 +420,7 @@ int quiche_conn_stream_shutdown(quiche_conn *conn, uint64_t stream_id,
                                 enum quiche_shutdown direction, uint64_t err);
 
 // Returns the stream's send capacity in bytes.
-ssize_t quiche_conn_stream_capacity(const quiche_conn *conn, uint64_t stream_id);
+ssize_t quiche_conn_stream_capacity(quiche_conn *conn, uint64_t stream_id);
 
 // Returns true if the stream has data that can be read.
 bool quiche_conn_stream_readable(const quiche_conn *conn, uint64_t stream_id);
@@ -485,6 +496,9 @@ void quiche_conn_peer_cert(const quiche_conn *conn, const uint8_t **out, size_t
 // Returns the serialized cryptographic session for the connection.
 void quiche_conn_session(const quiche_conn *conn, const uint8_t **out, size_t *out_len);
 
+// Returns the server name requested by the client.
+void quiche_conn_server_name(const quiche_conn *conn, const uint8_t **out, size_t *out_len);
+
 // Returns true if the connection handshake is complete.
 bool quiche_conn_is_established(const quiche_conn *conn);
 
@@ -548,6 +562,9 @@ typedef struct {
     // The number of QUIC packets that were lost.
     size_t lost;
 
+    // The number of QUIC packets that were marked as lost but later acked.
+    size_t spurious_lost;
+
     // The number of sent QUIC packets with retransmitted data.
     size_t retrans;
 
@@ -566,6 +583,12 @@ typedef struct {
     // The number of stream bytes retransmitted.
     uint64_t stream_retrans_bytes;
 
+    // The number of DATAGRAM frames received.
+    size_t dgram_recv;
+
+    // The number of DATAGRAM frames sent.
+    size_t dgram_sent;
+
     // The number of known paths for the connection.
     size_t paths_count;
 
@@ -580,6 +603,31 @@ typedef struct {
 
     // The number of streams stopped by remote.
     uint64_t stopped_stream_count_remote;
+
+    // The number of DATA_BLOCKED frames sent due to hitting the connection
+    // flow control limit.
+    uint64_t data_blocked_sent_count;
+
+    // The number of STREAM_DATA_BLOCKED frames sent due to a stream hitting
+    // the stream flow control limit.
+    uint64_t stream_data_blocked_sent_count;
+
+    // The number of DATA_BLOCKED frames received from the remote.
+    uint64_t data_blocked_recv_count;
+
+    // The number of STREAM_DATA_BLOCKED frames received from the remote.
+    uint64_t stream_data_blocked_recv_count;
+
+    // The total number of PATH_CHALLENGE frames that were received.
+    uint64_t path_challenge_rx_count;
+
+    // Total duration during which this side of the connection was
+    // actively sending bytes or waiting for those bytes to be acked.
+    uint64_t bytes_in_flight_duration_msec;
+
+    // True if the send buffer is in an inconsistent state, which could lead to
+    // connection stalls  or excess buffering.
+    bool tx_buffered_inconsistent;
 } quiche_stats;
 
 // Collects and returns statistics about the connection.
@@ -657,9 +705,32 @@ typedef struct {
     // The number of sent QUIC packets with retransmitted data on this path.
     size_t retrans;
 
+    // The number of times PTO (probe timeout) fired.
+    //
+    // Loss usually happens in a burst so the number of packets lost will
+    // depend on the volume of inflight packets at the time of loss (which
+    // can be arbitrary). PTO count measures the number of loss events and
+    // provides a normalized loss metric.
+    size_t total_pto_count;
+
+    /// The number of DATAGRAM frames received.
+    size_t dgram_recv;
+
+    /// The number of DATAGRAM frames sent.
+    size_t dgram_sent;
+
     // The estimated round-trip time of the path (in nanoseconds).
     uint64_t rtt;
 
+    // The minimum round-trip time observed (in nanoseconds).
+    uint64_t min_rtt;
+
+    // The maximum round-trip time observed (in nanoseconds).
+    uint64_t max_rtt;
+
+    // The estimated round-trip time variation (in nanoseconds).
+    uint64_t rttvar;
+
     // The size of the path's congestion window in bytes.
     size_t cwnd;
 
@@ -680,6 +751,13 @@ typedef struct {
 
     // The most recent data delivery rate estimate in bytes/s.
     uint64_t delivery_rate;
+
+    /// The maximum bandwidth estimate for the connection in bytes/s.
+    uint64_t max_bandwidth;
+
+    // The congestion window in bytes at the end of the startup or slow start,
+    // or 0 if the connection is still in startup.
+    uint64_t startup_exit_cwnd;
 } quiche_path_stats;
 
 
@@ -787,36 +865,36 @@ enum quiche_path_event_type {
 typedef struct quiche_path_event quiche_path_event;
 
 // Retrieves the next event. Returns NULL if there is no event to process.
-const quiche_path_event *quiche_conn_path_event_next(quiche_conn *conn);
+quiche_path_event *quiche_conn_path_event_next(quiche_conn *conn);
 
 // Returns the type of the event.
-enum quiche_path_event_type quiche_path_event_type(quiche_path_event *ev);
+enum quiche_path_event_type quiche_path_event_type(const quiche_path_event *ev);
 
 // Should be called if the quiche_path_event_type(...) returns QUICHE_PATH_EVENT_NEW.
-void quiche_path_event_new(quiche_path_event *ev,
+void quiche_path_event_new(const quiche_path_event *ev,
                            struct sockaddr_storage *local, socklen_t *local_len, struct sockaddr_storage *peer, socklen_t *peer_len);
 
 // Should be called if the quiche_path_event_type(...) returns QUICHE_PATH_EVENT_VALIDATED.
-void quiche_path_event_validated(quiche_path_event *ev,
+void quiche_path_event_validated(const quiche_path_event *ev,
                            struct sockaddr_storage *local, socklen_t *local_len, struct sockaddr_storage *peer, socklen_t *peer_len);
 
 // Should be called if the quiche_path_event_type(...) returns QUICHE_PATH_EVENT_FAILED_VALIDATION.
-void quiche_path_event_failed_validation(quiche_path_event *ev,
+void quiche_path_event_failed_validation(const quiche_path_event *ev,
                            struct sockaddr_storage *local, socklen_t *local_len, struct sockaddr_storage *peer, socklen_t *peer_len);
 
 // Should be called if the quiche_path_event_type(...) returns QUICHE_PATH_EVENT_CLOSED.
-void quiche_path_event_closed(quiche_path_event *ev,
+void quiche_path_event_closed(const quiche_path_event *ev,
                            struct sockaddr_storage *local, socklen_t *local_len, struct sockaddr_storage *peer, socklen_t *peer_len);
 
 // Should be called if the quiche_path_event_type(...) returns QUICHE_PATH_EVENT_REUSED_SOURCE_CONNECTION_ID.
-void quiche_path_event_reused_source_connection_id(quiche_path_event *ev, uint64_t *id,
+void quiche_path_event_reused_source_connection_id(const quiche_path_event *ev, uint64_t *id,
                            struct sockaddr_storage *old_local, socklen_t *old_local_len,
                            struct sockaddr_storage *old_peer, socklen_t *old_peer_len,
                            struct sockaddr_storage *local, socklen_t *local_len,
                            struct sockaddr_storage *peer, socklen_t *peer_len);
 
 // Should be called if the quiche_path_event_type(...) returns QUICHE_PATH_EVENT_PEER_MIGRATED.
-void quiche_path_event_peer_migrated(quiche_path_event *ev,
+void quiche_path_event_peer_migrated(const quiche_path_event *ev,
                            struct sockaddr_storage *local, socklen_t *local_len,
                            struct sockaddr_storage *peer, socklen_t *peer_len);
 
@@ -990,6 +1068,15 @@ enum quiche_h3_error {
 
     // See QUICHE_ERR_KEY_UPDATE.
     QUICHE_H3_TRANSPORT_ERR_KEY_UPDATE = QUICHE_ERR_KEY_UPDATE - 1000,
+
+    // See QUICHE_ERR_CRYPTO_BUFFER_EXCEEDED.
+    QUICHE_H3_TRANSPORT_ERR_CRYPTO_BUFFER_EXCEEDED = QUICHE_ERR_CRYPTO_BUFFER_EXCEEDED - 1000,
+
+    // See QUICHE_ERR_INVALID_ACK_RANGE.
+    QUICHE_H3_TRANSPORT_ERR_INVALID_ACK_RANGE = QUICHE_ERR_INVALID_ACK_RANGE - 1000,
+
+    // See QUICHE_ERR_OPTIMISTIC_ACK_DETECTED.
+    QUICHE_H3_TRANSPORT_ERR_OPTIMISTIC_ACK_DETECTED = QUICHE_ERR_OPTIMISTIC_ACK_DETECTED - 1000,
 };
 
 // Stores configuration shared between multiple connections.
@@ -1060,8 +1147,8 @@ int quiche_h3_for_each_setting(quiche_h3_conn *conn,
                                          uint64_t value, void *argp),
                                void *argp);
 
-// Check whether data will follow the headers on the stream.
-bool quiche_h3_event_headers_has_body(quiche_h3_event *ev);
+// Check whether more frames will follow the headers on the stream.
+bool quiche_h3_event_headers_has_more_frames(quiche_h3_event *ev);
 
 // Check whether or not extended connection is enabled by the peer
 bool quiche_h3_extended_connect_enabled_by_peer(quiche_h3_conn *conn);
@@ -1099,6 +1186,12 @@ int quiche_h3_send_response_with_priority(quiche_h3_conn *conn,
                             const quiche_h3_header *headers, size_t headers_len,
                             quiche_h3_priority *priority, bool fin);
 
+// Sends additional HTTP/3 headers on the specified stream.
+int quiche_h3_send_additional_headers(quiche_h3_conn *conn,
+                            quiche_conn *quic_conn, uint64_t stream_id,
+                            quiche_h3_header *headers, size_t headers_len,
+                            bool is_trailer_section, bool fin);
+
 // Sends an HTTP/3 body chunk on the given stream.
 ssize_t quiche_h3_send_body(quiche_h3_conn *conn, quiche_conn *quic_conn,
                             uint64_t stream_id, const uint8_t *body, size_t body_len,
@@ -1140,6 +1233,17 @@ int quiche_h3_take_last_priority_update(quiche_h3_conn *conn,
 bool quiche_h3_dgram_enabled_by_peer(quiche_h3_conn *conn,
                                      quiche_conn *quic_conn);
 
+typedef struct {
+    // The number of bytes received on the QPACK encoder stream.
+    uint64_t qpack_encoder_stream_recv_bytes;
+
+    // The number of bytes received on the QPACK decoder stream.
+    uint64_t qpack_decoder_stream_recv_bytes;
+} quiche_h3_stats;
+
+// Collects and returns statistics about the connection.
+void quiche_h3_conn_stats(const quiche_h3_conn *conn, quiche_h3_stats *out);
+
 // Frees the HTTP/3 connection object.
 void quiche_h3_conn_free(quiche_h3_conn *conn);
 
